use pyo3::prelude::*;

mod ripgrep_core;
use ripgrep_core::{PySortMode, PySortModeKind, py_files, py_search};

#[pymodule(gil_used = false)]
fn ripgrep_rs(m: &Bound<'_, PyModule>) -> PyResult<()> {
    m.add_class::<PySortMode>()?;
    m.add_class::<PySortModeKind>()?;
    m.add_function(wrap_pyfunction!(py_search, m)?)?;
    m.add_function(wrap_pyfunction!(py_files, m)?)?;
    Ok(())
}
